import json

from mcclient.response import SLPResponse, LegacySLPResponse
from mcclient.base_client import BaseClient
from mcclient.base_client import DEFAULT_PORT, DEFAULT_TIMEOUT, DEFAULT_PROTO
from mcclient.encoding.packet import Packet


class SLPClient(BaseClient):
    def __init__(self, host: str, port: int = DEFAULT_PORT, timeout: int = DEFAULT_TIMEOUT, proto: int = DEFAULT_PROTO,
                 srv: bool = True):
        super().__init__(host=host, port=port, timeout=timeout, proto=proto, srv=srv)

    def get_status(self) -> SLPResponse:
        self._connect()
        self._handshake()
        res = self._status_request()
        return SLPResponse(self.address.get_host(), self.address.get_port(), res)

    def _status_request(self) -> dict:
        packet = Packet(b"\x00")  # send status request
        self._send(packet)
        _, res = self._recv()
        self._close()

        # TODO: read string length from bytes
        res_data = res[2:]
        res_str = res_data.decode("utf-8")
        res_dict = json.loads(res_str)
        return res_dict


class LegacySLPClient(BaseClient):
    def __init__(self, host: str, port: int = 25565, timeout: int = 5):
        super().__init__(host=host, port=port, timeout=timeout)

    def get_status(self) -> LegacySLPResponse:
        self._connect()
        # legacy status request
        self.sock.send(b"\xFE\x01")
        raw_res = self.sock.recv(1024)
        self._close()

        # remove padding and other headers
        res_bytes = raw_res[3:]
        res_str = res_bytes.decode("UTF-16-be", errors="ignore")
        res_split_str = res_str.split("\x00")
        res = LegacySLPResponse(self.address.get_host(), self.address.get_port(), res_split_str)
        return res



import datetime
import re


class Players:
    online: int
    max: int | None
    list: list | None

    def __init__(self, online: int, max: int | None, player_list: list | None):
        self.online = online
        self.max = max
        self.list = player_list


class Version:
    name: str
    protocol: int | None
    software: str
    brand: str

    def __init__(self, name: str, protocol: int | None):
        self.name = name
        self.protocol = protocol


class StatusResponse:
    host: str
    port: int
    raw_res: str | dict | list
    res: dict
    timestamp: str

    def __init__(self, host: str, port: int, raw_res: str | dict | list):
        self.host = host
        self.port = port
        self.raw_res = raw_res
        self.res = {"host": self.host, "port": self.port}
        self.timestamp = str(datetime.datetime.now())

    @staticmethod
    def _remove_color_codes(cstr: str, bedrock: bool = False) -> str:
        """
        Returns the input string stripped of all Minecraft color/formatting codes.
        documentation on minecraft color/formatting codes:
        https://minecraft.fandom.com/wiki/Formatting_codes

        Args:
        cstr (str): The string to remove color codes from.
        bedrock (bool): Wether to method should remove bedrock color/formating codes.

        Returns:
        str: The input string stripped of all Minecraft color/formatting codes.
        """

        if bedrock:
            return re.sub(r"§[a-u0-9]", "", cstr)
        return re.sub(r"§[a-fk-or0-9]", "", cstr)


class SLPResponse(StatusResponse):
    motd: str | dict
    favicon: str | None
    version: Version
    players: Players

    def __init__(self, host: str, port: int, raw_res: dict):
        super().__init__(host, port, raw_res)

        self.res = self.res | self._parse_slp_res(self.raw_res)
        self.motd = self.res["motd"]
        self.favicon = self.res["favicon"]
        self.version = Version(
            self.res["version"]["name"], self.res["version"]["protocol"])
        self.players = Players(
            self.res["players"]["online"], self.res["players"]["max"], self.res["players"]["list"])

    def _parse_slp_res(self, slp_res) -> dict:
        slp_res = self._add_missing(slp_res)

        if "sample" in slp_res["players"]:
            slp_res["players"]["list"] = slp_res["players"]["sample"]
            slp_res["players"].pop("sample")

        if "description" in slp_res:
            slp_res["motd"] = slp_res["description"]
            slp_res.pop("description")

        else:
            slp_res["motd"] = ""

        for player in slp_res["players"]["list"]:
            player["last_seen"] = self.timestamp

        if "favicon" in slp_res:
            slp_res["favicon"] = slp_res["favicon"]

        else:
            slp_res["favicon"] = False

        slp_res["motd"] = self._parse_motd(slp_res["motd"])
        slp_res["version"]["name"] = self._remove_color_codes(
            cstr=slp_res["version"]["name"])
        slp_res["status"] = "online"
        return slp_res

    @classmethod
    def _parse_motd(cls, raw_motd: str) -> str:
        motd = ""
        if isinstance(raw_motd, dict):
            entries = raw_motd.get("extra", [])
            end = raw_motd.get("text", "")

            for entry in entries:
                motd += entry.get("text", "")
            motd += end

        elif isinstance(raw_motd, str):
            motd = raw_motd

        motd = motd.replace("\n", " ").strip()
        motd = cls._remove_color_codes(cstr=motd)
        return motd

    @staticmethod
    def _add_missing(res: dict) -> dict:
        default_res: dict = {
            "previewsChat": None,
            "enforcesSecureChat": None,
            "description": "",
            "version": {"name": "", "protocol": -1},
            "players": {"online": -1, "max": -1, "list": []}
        }

        for key in default_res:
            if key not in res:
                res[key] = default_res[key]

        res["players"]["list"] = default_res["players"]["list"]
        return res


class LegacySLPResponse(StatusResponse):
    motd: str
    version: Version
    players: Players

    def __init__(self, host, port, raw_res):
        super().__init__(host, port, raw_res)

        self.res = self.res | self._parse_res(self.raw_res)
        self.motd = self.res["motd"]
        self.version = Version(self.res["version"], None)
        self.players = Players(self.res["online"], self.res["max"], None)

    @staticmethod
    def _parse_res(raw_res) -> dict:
        res = {"version": raw_res[2], "motd": raw_res[3], "online": raw_res[4], "max": raw_res[5]}
        return res


class QueryResponse(StatusResponse):
    motd: str
    gametype: str
    game_id: str
    plugins: list
    map: str
    hostip: str
    hostport: int
    players: Players
    version: Version

    def __init__(self, host: str, port: int, raw_res: dict):
        super().__init__(host, port, raw_res)

        self.res = raw_res | self.res

        self.motd = self.res["motd"]
        self.gametype = self.res["gametype"]
        self.game_id = self.res["game_id"]
        self.plugins = self.res["plugins"]
        self.map = self.res["map"]
        self.hostip = self.res["hostip"]
        self.hostport = self.res["hostport"]

        self.players = Players(
            self.res["numplayers"], self.res["maxplayers"], self.res["players"])
        self.version = Version(self.res["version"], None)
        self.version.software = self.res["software"]


class BedrockResponse(StatusResponse):
    version: Version
    motd: str
    players: Players
    server_id: int
    map: str
    gametype: str

    def __init__(self, host: str, port: int, raw_res: list):
        super().__init__(host, port, raw_res)

        self.res = {"version": {
            "brand": self.raw_res[0],
            "protocol": int(self.raw_res[2]),
            "name": self.raw_res[3]
        }, "motd": self._remove_color_codes(cstr=self.raw_res[1], bedrock=True), "online_players": int(self.raw_res[4]),
            "max_players": int(self.raw_res[5]), "server_id": self.raw_res[6], "map": None, "gametype": None}

        if len(self.raw_res) > 6:
            self.res["map"] = self.raw_res[7]

        if len(self.raw_res) > 7:
            self.res["gametype"] = self.raw_res[8]

        self.version = Version(
            self.res["version"]["name"], self.res["version"]["protocol"])
        self.version.brand = self.res["version"]["brand"]
        self.motd = self.res["motd"]
        self.players = Players(self.res["online_players"], self.res["max_players"], None)
        self.server_id = self.res["server_id"]
        self.map = self.res["map"]
        self.gametype = self.res["gametype"]



import socket
import struct

from mcclient.address import Address
from mcclient.encoding.packet import Packet
from mcclient.encoding.varint import pack_varint, read_varint

DEFAULT_HOST = "localhost"
DEFAULT_PORT = 25565
DEFAULT_TIMEOUT = 5
DEFAULT_PROTO = 47


class IncompletePacket(Exception):
    def __init__(self, size, missing):
        super().__init__(f"Incomplete packet: missing {missing} from {size} bytes.")


class BaseClient:
    address: Address
    sock: socket.socket
    connected: bool
    protocol_version: int

    def __init__(self, host: str = DEFAULT_HOST, port: int = DEFAULT_PORT, timeout: int = DEFAULT_TIMEOUT,
                 proto: int = DEFAULT_PROTO, srv: bool = True) -> None:
        self.address = Address(host, port)
        self.connected = False
        self.protocol_version = proto
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.settimeout(timeout)
        if srv:
            self.address.resolve()

    def implant_socket(self, sock: socket.socket) -> None:
        self.sock = sock
        self.connected = True

    def _connect(self) -> None:
        if self.connected:
            return

        self.sock.connect(self.address.address())
        self.connected = True

    def _handshake(self, next_state: int = 1) -> None:
        packet = Packet(
            b"\x00",  # packet id
            pack_varint(self.protocol_version),
            self.address.get_host(),
            struct.pack(">H", self.address.get_port()),
            pack_varint(next_state)  # next state 1 for status request
        )
        self._send(packet)

    def _send(self, packet: Packet) -> int:
        return self.sock.send(packet.pack())

    def _recv(self) -> tuple[int, bytes]:
        length = read_varint(self.sock)
        packet_id = read_varint(self.sock)
        return packet_id, self._recv_bytes(length)

    def _recv_bytes(self, length: int) -> bytes:
        received = 0
        data = b""
        while received < length - len(pack_varint(length)):
            chunk = self.sock.recv(length - received)
            data += chunk
            received += len(chunk)

            if chunk == b"":
                raise IncompletePacket(length, length - received)

        return data

    def _close(self):
        self.sock.close()
        self.connected = False



import struct
import socket


def pack_varint(data: int) -> bytes:
    ordinal = b''
    while data != 0:
        byte = data & 0x7F
        data >>= 7
        ordinal += struct.pack('B', byte | (0x80 if data > 0 else 0))
    return ordinal


def read_varint(sock: socket.socket) -> int:
    data = 0
    for i in range(5):
        ordinal = sock.recv(1)
        if len(ordinal) == 0:
            break

        byte = ord(ordinal)
        data |= (byte & 0x7F) << 7 * i
        if not byte & 0x80:
            break
    return data



import socket
import struct
import random

from mcclient.response import QueryResponse
from mcclient.encoding.packet import QueryPacket
from mcclient.base_client import DEFAULT_PORT, DEFAULT_TIMEOUT


class QueryClient:
    host: str
    port: int
    sock: socket.socket
    session_id: int
    token: bytes

    def __init__(self, host: str, port: int = DEFAULT_PORT, timeout: int = DEFAULT_TIMEOUT):
        self.host = host
        self.port = port
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.settimeout(timeout)

    def _handshake(self) -> None:
        # generate session id from an integer between 0 and 2147483648
        self.session_id = random.randint(0, 2147483648) & 0x0F0F0F0F
        packet = QueryPacket(
            9,  # type 9 for handshaking
            self.session_id,
            b""  # empty payload
        )
        self._send(packet)
        res = self._recv()
        # extract token from response
        self.token = struct.pack('>l', int(res[2][:-1]))

    def _send(self, packet: QueryPacket) -> None:
        self.sock.sendto(packet.pack(), (self.host, self.port))

    def _recv(self) -> tuple[int, bytes, bytes]:
        res = self.sock.recv(4096)
        type = res[0]
        session_id = res[1:5]
        payload = res[5:]
        return type, session_id, payload

    def _query_request(self) -> dict:
        self._handshake()
        # challenge token and some padding for a full status request
        payload = self.token + b"\x00\x00\x00\x00"
        packet = QueryPacket(
            0,  # 0 for a status request
            self.session_id,
            payload
        )
        self._send(packet)
        raw_res = self._recv()
        return self._read_query(raw_res[2])

    def get_status(self) -> QueryResponse:
        res = self._query_request()
        return QueryResponse(self.host, self.port, res)

    @staticmethod
    def _read_query(res: bytes) -> dict:
        # remove padding
        res = res[11:]

        # split stats from players
        raw_stats, raw_players = res.split(b"\x00\x00\x01player_\x00\x00")

        raw_split_stats = raw_stats.split(b"\x00")
        # decode keys and values
        stats = [stat.decode("utf-8")
                 for stat in raw_split_stats]

        # replace "hostname" key with "motd" key
        stats[0] = "motd"
        key_field = True
        data: dict = {}
        for y, x in enumerate(stats):
            if key_field:
                data[x] = stats[y + 1]
                key_field = False

            else:
                key_field = True

        # convert strings to ints
        for key in ["numplayers", "maxplayers", "hostport"]:
            data[key] = int(data[key])

        data["software"] = "Vanilla"
        software_parts = data["plugins"].split(":", 1)
        data["software"] = software_parts[0].strip()
        if len(software_parts) == 2:
            data["plugins"] = [plugin.strip()
                               for plugin in software_parts[1].split(";")]

        else:
            data["plugins"] = []

        # remove end padding
        raw_players = raw_players[:-2]
        # split players
        players = raw_players.split(b"\x00")
        # split players
        data["players"] = [player.decode(
            "utf-8") for player in players if player != b""]
        return data



from mcclient import LegacySLPClient

from tests.utils import TooManyPackets, BaseTestConn, create_mock_socket

TEST_HOSTNAME = "example.com"


class LegacySLPTestConn(BaseTestConn):
    max_packets = 1

    def __init__(self):
        super().__init__()

    def send(self, data):
        if self.packets == 0:
            self.packets += 1
            assert data == b"\xFE\x01"

            test_res = b"""\xFF\x00\x25\x00\xA7\x00\x31\x00\x00\x00
                \x31\x00\x32\x00\x37\x00\x00\x00\x31\x00\x2E\x00\x31
                \x00\x39\x00\x2E\x00\x33\x00\x00\x00\x41\x00\x20\x00
                \x4D\x00\x69\x00\x6E\x00\x65\x00\x63\x00\x72\x00\x61
                \x00\x66\x00\x74\x00\x20\x00\x53\x00\x65\x00\x72\x00
                \x76\x00\x65\x00\x72\x00\x00\x00\x30\x00\x00\x00\x32
                \x00\x30"""
            # test packet dump from a raw legacy slp response, SHOULD be changed in the future!!
            self.respond(test_res)

        elif self.packets >= self.max_packets:
            raise TooManyPackets(self.max_packets)


def test_legacy_slp_request():
    socket = create_mock_socket(LegacySLPTestConn)
    test_conn = socket.socket()
    legacy_slp_client = LegacySLPClient(TEST_HOSTNAME)
    legacy_slp_client.implant_socket(test_conn)
    legacy_slp_client.get_status()



import struct

from mcclient.encoding.varint import pack_varint


class Packet:
    fields: tuple

    def __init__(self, *fields):
        self.fields = fields

    def pack(self) -> bytes:
        packet = b""
        for field in self.fields:
            packet += self._encode(field)

        # add the packet length
        return pack_varint(len(packet)) + packet

    @staticmethod
    def _encode(data) -> bytes:
        if isinstance(data, str):
            str_bytes = data.encode("utf-8")
            return pack_varint(len(str_bytes)) + str_bytes

        elif isinstance(data, bytes):
            return data

        elif isinstance(data, bool):
            return b"\x01" if data else b"\x00"

        raise TypeError(f"Type {type(data)} cannot be encoded")


class QueryPacket:
    type: int
    session_id: int
    payload: bytes

    def __init__(self, type: int, session_id: int, payload: bytes):
        self.type = type
        self.session_id = session_id
        self.payload = payload

    def pack(self) -> bytes:
        packet = b"\xFE\xFD"  # query packet padding
        packet += struct.pack("!B", self.type)
        packet += struct.pack('>l', self.session_id)
        packet += self.payload
        return packet



import ipaddress

import dns.resolver


class Address:
    host: str
    port: int
    srv_host: str
    srv_port: int
    srv: bool

    def __init__(self, host: str, port: int) -> None:
        self.host = host
        self.port = port
        self.srv = False

    def resolve(self) -> None:
        if self.is_ip():
            return

        try:
            record = dns.resolver.resolve("_minecraft._tcp." + self.host, "SRV")[0]
            self.srv_host = str(record.target).rstrip(".")
            self.srv_port = record.port
            self.srv = True

        except Exception:
            pass

    def address(self) -> tuple[str, int]:
        return self.get_host(), self.get_port()

    def get_host(self) -> str:
        if self.srv:
            return self.srv_host

        return self.host

    def get_port(self) -> int:
        if self.srv:
            return self.srv_port

        return self.port

    def is_ip(self) -> bool:
        try:
            ipaddress.ip_address(self.host)
            return True

        except ValueError:
            return False



import json
import struct

from mcclient import SLPClient
from mcclient.encoding.packet import Packet
from mcclient.encoding.varint import pack_varint

from tests.utils import TooManyPackets, BaseTestConn, create_mock_socket

TEST_PROTO = 47
TEST_HOSTNAME = "example.com"
TEST_PORT = 25565
BASE_RES = {
    "description": {
        "text": "This is an example motd"
    },
    "version": {
        "name": "1.19.2",
        "protocol": 795
    },
    "players": {
        "max": 69,
        "online": 5,
        "sample": [
            {
                "name": "thinkofdeath",
                "id": "4566e69f-c907-48ee-8d71-d7ba5aa00d20"
            }
        ]
    },
    "previewsChat": True,
    "enforcesSecureChat": True
}


class SLPTestConn(BaseTestConn):
    max_packets = 2

    def __init__(self, hostname, base_res, proto):
        super().__init__()
        self.hostname = hostname
        self.base_res = base_res
        self.proto = proto

    def send(self, data):
        if self.packets == 0:
            self.packets += 1
            test_packet = Packet(
                b"\x00",  # packet id
                pack_varint(self.proto),
                self.hostname,
                struct.pack(">H", TEST_PORT),
                pack_varint(1)
            )
            assert data == test_packet.pack()

        elif self.packets == 1:
            self.packets += 1
            assert data == b"\x01\x00"
            self.respond_res()

        elif self.packets >= self.max_packets:
            raise TooManyPackets(self.max_packets)

    def respond_res(self):
        packet = Packet(
            b"\x00",
            json.dumps(self.base_res)
        )
        self.respond(packet.pack())


def test_slp_request():
    socket = create_mock_socket(SLPTestConn)
    test_conn = socket.socket(TEST_HOSTNAME, BASE_RES, TEST_PROTO)
    slp_client = SLPClient(TEST_HOSTNAME, proto=TEST_PROTO)
    slp_client.implant_socket(test_conn)
    slp_client.get_status()



from mcclient.address import Address


def test_address():
    # random addresses for testing

    assert Address("google.com", 0).get_host() == "google.com"
    # add more tests for test_domain

    assert Address("23.23.23.23", 0).get_host() == "23.23.23.23"

    # random addresses for testing
    addr = Address("hypixel.net", 0)
    addr.resolve()
    assert addr.address() == ("mc.hypixel.net", 25565)



import socket
import struct

from mcclient.address import Address
from mcclient.response import BedrockResponse
from mcclient.base_client import DEFAULT_TIMEOUT

DEFAULT_BEDROCK_PORT = 19132


class BedrockSLPClient:
    address: Address
    sock: socket.socket

    def __init__(self, host: str, port: int = DEFAULT_BEDROCK_PORT, timeout: int = DEFAULT_TIMEOUT):
        self.address = Address(host, port)
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.settimeout(timeout)

    def get_status(self) -> BedrockResponse:
        raw_res = self._request_status()
        res = self._parse_res(raw_res)
        return BedrockResponse(self.address.host, self.address.port, res)

    def _request_status(self) -> bytes:
        """
        needs to be updated (https://wiki.vg/Raknet_Protocol#Unconnected_Ping)
        """
        status_request = b"\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff\x00\xfe\xfe\xfe\xfe\xfd\xfd\xfd\xfd\x124Vx"
        self.sock.sendto(status_request, self.address.address())
        return self.sock.recv(4096)

    @staticmethod
    def _parse_res(raw_bytes: bytes) -> list[str]:
        res_bytes = raw_bytes[1:]
        extra_len = struct.unpack(">H", res_bytes[32:34])[0]
        res_bytes = res_bytes[34: 34 + extra_len]
        res_str = res_bytes.decode()
        res_split = res_str.split(";")
        return res_split
